/* schedprt */
#define VERSION "1.2"

/*  V1.0 Initial release at the end of 2006.
    V1.1 - Jul 2007 Fixed bug where songs that exactly fitted one page of the Generic/Text
	Only Printer caused the author and copyright information to be appended to the
	song. A chunk of subroutine get_song was re-written.
    V1.2 - 7 Jan 2012 added a running tally of song to help with diagnostics.
*/

/* Uncomment the next line or use -DUNIX to compile on a unix system */
/* #define UNIX */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define NULLCH (char)0

void usage(void)
{
	fprintf(stderr,"Usage (v%s): schedprt [-c config.dat] [-o output.rtf] <filename.txt> -[h|?]\n",
		VERSION);
}

/* Copyright (c) 2006-07 Rick Parsons.

	This software is published Open Source Freeware. You are free to use the
	programs either in full or part without charge. I would ask however that
	this notice is included with any extract or onward distribution.

	No charge may be made for the use of this code, however there is no
	objection to including modest extracts (with acknowledgement) in otherwise
	commercial software.

	No warranty is given or implied by the use of the software which you do at
	your own risk. No compensation can be considered regarding damage to data,
	computers or any thing else arising as a result of using these programs.
*/

#define TRUE 1
#define FALSE 0

int options(int, char **);
void init(void);
void get_template(FILE *);
void get_tname(FILE *, char *);
void get_word(FILE *, char *);
int get_line(FILE *, char *);
void get_escape(FILE *, char *);
void get_schedule(FILE *);
char * get_song(FILE *);
char * get_stanza(void);
void put_schedule(FILE *, FILE *);
void put_text(FILE *, char *, FILE *);
char * store_escape(int, char *);
int is_tag(char *);

void splitpath(const char *, char *, char *, char *);
char * makepath(const char *, const char *, const char *);

/* escape = an array of char *, one for each possible escape a-z, A-Z, 0-9
		(assumes numbers before upper case before lower case)
   output = output file name
   config = config file name
   songbuf = buffer for one song
   temp = a buffer for various purposes
   inschd = a flag which is set while processing schedule content
   insong = a flag which is set while processing song content
	These are to guard against some indefinate loops
   sbp = a pointer into songbuf to mark the next stanza
   tags = a table of valid stanza tags
*/

#define ESCAPECHARS 'z'-'0'+1
static char * escape[ESCAPECHARS];
static char output[128];
static char config[128];
#define SONGBUFSZ 2048
static char songbuf[SONGBUFSZ];
static char temp[SONGBUFSZ];
static int inschd, insong;
static char * sbp;
static char * tags[] = {"Verse", "Chorus", "Bridge", "Intro", "End", "Slide", "Tag", ""};

/* fixed meaning escape codes
	: = Schedule template
	; = song template
	< = end of line template
	0 = stanza template
	1-7 = tag templates as above
	A = Author
	C = schedule Content
	L = Licence
	N = song Number
	S = Schedule name
	T = Title
	a = administrator data
	c = copyright
	n = tag text
	s = song content
	t = tag
	z = stanza content
	the rest in the gaps not used
*/
#define ESCAPE_schedule ':'-'0'
#define ESCAPE_song ';'-'0'
#define ESCAPE_line '<'-'0'
#define ESCAPE_stanza '0'-'0'
#define ESCAPEA 'A'-'0'
#define ESCAPEC 'C'-'0'
#define ESCAPEL 'L'-'0'
#define ESCAPEN 'N'-'0'
#define ESCAPES 'S'-'0'
#define ESCAPET 'T'-'0'
#define ESCAPEa 'a'-'0'
#define ESCAPEc 'c'-'0'
#define ESCAPEn 'n'-'0'
#define ESCAPEs 's'-'0'
#define ESCAPEt 't'-'0'
#define ESCAPEz 'z'-'0'

#define BUFSZ 1024

int _main(int argc, char *argv[])
{
	extern char output[];
	extern char config[];
	int i;
	char input[128]; /* input file name */
	char path[128];
	char fname[128];
	char ext[128];
	FILE * in;
	FILE * out;
	FILE * cfg;

	/* preset some variables */
	*output = NULLCH;
	*config = NULLCH;

	/******************************/
	/* Decode the call parameters */
	/******************************/

	if (argc == 1)
	{
		/* If no arguments we call the usage routine and exit */
		usage();
		return 1;
	}

	/* handle the program options */
	i = options(argc,argv);

	/* If there is still no file name then what do we do? */
	if (i == 0)
	{
		usage();
		return 2;
	}

	/* check bounds of input path before using */
	if (strlen(argv[i]) > 127)
	{
		fprintf(stderr, "Fatal: File path too long, move it closer to the root of the disk\n");
		return 12;
	}

	strcpy(input, argv[i]);

	if (*output == NULLCH)
	{
		/* if we have no output file name then create it from the input */
		splitpath(input, path, fname, ext);
		strcpy(output, makepath(path, fname, "rtf"));
	}

	if (*config == NULLCH)
	{
		/* if we have no config file name then use the default */
		/* look for it in the program path, not cwd */
		/* check size of path before using */
		if (strlen(argv[0]) > 126)
		{
			fprintf(stderr, "Fatal: Program path too long, move it closer to the root of the disk\n");
			return 12;
		}
		splitpath(argv[0], path, fname, ext);
		strcpy(config, makepath(path, "schedprt", "dat"));
	}

	/* initialise some stuff, mostly the escape array */
	init();

	/* open the configuration file */
	if ((cfg = fopen(config, "r")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Configuration file %s failed to open\n", config);
		perror("Fatal");
		return 12;
	}

	/* read the configuration file */
	get_template(cfg);
	fclose(cfg);

	/* open the input file */
	if ((in = fopen(input, "r")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Input file %s failed to open\n", input);
		perror("Fatal");
		return 12;
	}

	/* open the output file */
	if ((out = fopen(output, "w")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Output file %s failed to open\n", output);
		perror("Fatal");
		fclose(in);
		return 12;
	}

	/* read the start of the data file to get the basic information */
	get_schedule(in);

	/* This is where the bulk of the work is done */
	put_schedule(out, in);

	fclose(out);
	fclose(in);

	return 0;
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash (unix) or a slash (DOS) */
int options(int argc,char *argv[])
{
	extern char output[];
	extern char config[];
	int i,firstnonoption=0;

	for (i=1; i < argc;i++)
	{
		if (
#ifndef UNIX
			argv[i][0] == '/' ||
#endif
			argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
				case 'c':
				case 'C': /* Configuration file name */
					strcpy(config, argv[++i]);
					break;
				case 'o':
				case 'O': /* Output file name */
					strcpy(output, argv[++i]);
					break;
				case 'h':
				case 'H':
				case '?':
					usage();
					return 0;
				default:
					fprintf(stderr,"Unknown option %s\n",argv[i]);
					break;
			}
		}
		else
		{
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

void splitpath(const char * string, char * path, char * file, char * ext)
{
	char * cp;

	strcpy(path, string);
	*file = NULLCH;
	*ext = NULLCH;

#ifdef UNIX
	if ((cp = strrchr(path, '/')) != (char *)NULL)
#else
	if ((cp = strrchr(path, '\\')) != (char *)NULL)
#endif
	{
		*cp = NULLCH;
		strcpy(file, cp + 1);
	}
	else
	{
		strcpy(file, path);
		*path = NULLCH;
	}

	if ((cp = strchr(file, '.')) != (char *)NULL)
	{
		*cp = NULLCH;
		strcpy(ext, cp + 1);
	}

	return;
}

char * makepath(const char * path, const char * file, const char * ext)
{
	extern char temp[];

	*temp = NULLCH;
	if (*path != NULLCH)
	{
		strcpy(temp, path);
#ifdef UNIX
		strcat(temp, "/");
#else
		strcat(temp, "\\");
#endif
	}
	strcat(temp, file);
	strcat(temp, ".");
	strcat(temp, ext);

	return temp;
}

void init(void)
{
	extern char * escape[];
	int i;

	/* initialise all escapes to point to a null string */
	for (i = 0; i < ESCAPECHARS; i++)
	{
		escape[i] = (char *)NULL;
		store_escape(i, "");
	}

	/* One has a pre set value */
	store_escape(ESCAPE_line, "\\line\n");

	return;
}

void get_template(FILE * cfg)
{
	extern char * escape[];
	int i, esc;
	char c;

	/* fprintf(stderr, "get_template\n"); */

	while ((c = getc(cfg)) != EOF)
	{
		if (c != '%')
		{	/* Comment or other unknown text */
			if (c == '\n')
				continue;

			while ((c = getc(cfg)) != EOF)
				if (c == '\n')
					break;
			continue;
		}
		get_tname(cfg, temp);
		if (strcasecmp(temp, "Schedule") == 0)
			esc = ESCAPE_schedule;
		else if (strcasecmp(temp, "Song") == 0)
			esc = ESCAPE_song;
		else if (strcasecmp(temp, "Line") == 0)
			esc = ESCAPE_line;
		else if (strcasecmp(temp, "Stanza") == 0)
			esc = ESCAPE_stanza;
		else
		{
			/* see if this is a special tag template */
			for (i = 0; *tags[i] != NULLCH; i++)
			{
				if (strcasecmp(temp, tags[i]) == 0)
					break;
			}
			if (*tags[i] != NULLCH)
				esc = i + 1;
			else
			{
				/* Not one of these, try for a simple variable */
				if (*temp >= 'A' && *temp <= 'z')
					esc = *temp -'0';
				/* if we get here it is an unknown escape, ignore it */
				else
					continue;
			}
		}

		get_escape(cfg, temp);
		/* fprintf(stderr, "get_template type %c\n", esc + '0'); */
		store_escape(esc, temp);
	}

	return;
}

void get_tname(FILE * cfg, char * result)
{
	char * cp = result;
	char c;

	while ((c = getc(cfg)) != EOF)
	{
		if (c == '\n')
			break;
		*(cp++) = c;
	}
	*cp = NULLCH;
	/* fprintf(stderr, "get_tname returns %s\n", result); */
	return;
}

void get_escape(FILE * cfg, char * result)
{
	char * cp = result;
	char c;
	int cc;

	while ((c = getc(cfg)) != EOF)
	{
		if (c == '\n')
			cc = -1;
		else if (c == '%' && cc == 0)
		{
			if ((c = getc(cfg)) == '\\')
				break;
			*(cp++) = '%';
			cc++;
		}
		*(cp++) = c;
		cc++;
	}
	*(cp - 1) = NULLCH;
	/* fprintf(stderr, "get_escape returns %s\n", result); */
	return;
}

void get_word(FILE * in, char * result)
{
	char * cp = result;
	char c;

	/* skip over leading blank lines and spaces */
	while ((c = getc(in)) != EOF)
	{
		if (c != ' ' && c != '\n')
			break;
	}

	/* now extract a word ending with a space or newline */
	*(cp++) = c;
	while ((c = getc(in)) != EOF)
	{
		if (c == '\n' || c == ' ')
			break;
		*(cp++) = c;
	}
	*cp = NULLCH;
	/* fprintf(stderr, "get_word returns %s\n", result); */
	return;
}

int get_line(FILE * in, char * result)
{
	char * cp = result;
	char c;
	int i;

	/* skip over 5 leading spaces */
	for (i = 0; i <= 5; i++)
	{
		c = getc(in);
		if (c != ' ')
			break;
	}

	/* now extract a line ending with a newline or FF */
	while (c != '\n' && c != '\f')
	{
		*(cp++) = c;
		if ((c = getc(in)) == EOF)
		{
			--cp;
			break;
		}
	}

	*cp = NULLCH;
	/* fprintf(stderr, "get_line returns (%s) %d\n", result, c); */
	return c;
}

void get_schedule(FILE * in)
{
	char c;
	char * cp;

	/* fprintf(stderr, "get_schedule\n"); */

	get_word(in, temp);
	if (strcmp(temp, "Schedule") != 0)
		fprintf(stderr, "Warning: this doesn't look like a Schedule print.\n");
	get_word(in, temp);
	if (strcmp(temp, "Detail") != 0)
		fprintf(stderr, "Warning: this doesn't look like a Schedule Detail print.\n");

	/* the is the date if we want it */
	get_word(in, temp);

	/* the is the time if we want it */
	get_word(in, temp);

	/* skip over spaces */
	while ((c = getc(in)) != EOF)
	{
		if (c != ' ')
			break;
	}

	/* The schedule name starts with "for" and ends in .ews */
	cp = temp;
	*(cp++) = c;
	while ((c = getc(in)) != EOF)
	{
		if (c == '.')
			break;
		*(cp++) = c;
	}
	*cp = NULLCH;

	(void)store_escape(ESCAPES, temp + 3);

	/* skip to next line ready for songs */
	(void)get_line(in, temp);

	return;
}

char * get_song(FILE * in)
{
	extern char * escape[];
	extern char songbuf[];
	extern char * sbp;
	char * cp;
	char c;
	char temp2[SONGBUFSZ];

	/* fprintf(stderr, "get_song\n"); */
	*temp2 = NULLCH;

	/* Get the song title */
	c = get_line(in, temp);
	while (c != EOF && *temp == NULLCH)
		c = get_line(in, temp);
	(void)store_escape(ESCAPET, temp);

	/* if there wasn't one then we have done the last song */
	if (*temp == NULLCH)
		return escape[ESCAPET];

	/* v1.2 Print a running tally to stderr to reassure the user */
	fprintf(stderr, "%s\n", temp);

	/* Now get the song text */
	sbp = songbuf + 1;
	*sbp = NULLCH;
	(void)get_line(in, temp);
	while (*temp != NULLCH)
	{
		/* try to detect long songs that go over two pages.
		They stop dead at the end of a page, then there is a
		Licence line (with no author) and a Page number line
		some blank lines then another title line and it starts
		again */
		if (strncmp(temp, "                    ", 20) == 0)
		{
			(void)get_line(in, temp); /* the page number */
			c = get_line(in, temp); /* the blank lines */
			while (c != EOF && *temp == NULLCH)
				c = get_line(in, temp); /* the title line */
			(void)get_line(in, temp); /* the next song line */
			continue;
		}
		/* another possibility is a song of exactly one page. We will
		have overshot the end of the song and put the licence line
		into the lyrics (no way to spot that) but now we will have
		a line with the page number in it. */
		else if (strlen(temp) > 65 && strncmp(temp + 62, "Page", 4) == 0)
		{
			/* back off the last line of the song */
			--sbp;
			while (*(sbp - 1) != '\n')
				--sbp;
			/* save it */
			strcpy(temp2, temp);
			strcpy(temp, sbp);
			sbp++;
			break;
		}
		strcpy(sbp, temp);
		sbp += strlen(temp);
		*(sbp++) = '\n';
		(void)get_line(in, temp);
	}
	*(sbp - 1) = NULLCH;

	/* point to the start of the song.
	This little fiddle is to allow for empty songs - e.g. PowerPoint files */
	sbp = songbuf + 1;

	/* now skip over the blank lines until we get to the footer */
	while (*temp == NULLCH)
		(void)get_line(in, temp);

	/* is there an author field */
	if (strlen(temp) < 20)
	{
		(void)store_escape(ESCAPEA, "");
		cp = temp - 1;
	}
	else
	{
		/* the next bit relies on the author field not having two consecutive spaces */
		for (cp = temp; *cp != ' ' || *(cp + 1) != ' '; cp++)
			;
		*cp = NULLCH;
		(void)store_escape(ESCAPEA, temp);
	}

	/* Now hunt for the licence */
	while (*(++cp) == ' ')
		;
	(void)store_escape(ESCAPEL, cp);

	/* check to see that we don't have the next line saved */
	if (*temp2 == NULLCH)
		(void)get_line(in, temp);
	else
		strcpy (temp,temp2);

	/* is there a copyright field */
	if (strlen(temp) < 20)
		*temp = NULLCH;
	else
	{
		/* the next bit relies on the copyright field not having two consecutive spaces */
		for (cp = temp; *cp != ' ' || *(cp + 1) != ' '; cp++)
			;
		*cp = NULLCH;
	}
	(void)store_escape(ESCAPEc, temp);

	/* If the copyright is null then we don't need the licence either */
	if (*escape[ESCAPEc] == NULLCH)
		(void)store_escape(ESCAPEL, "");

	/* fprintf(stderr, "get_song returns %s\n", escape[ESCAPET]);
	fprintf(stderr, "songbuf has %s\n", songbuf + 1); */

	return escape[ESCAPET];
}

char * get_stanza(void)
{
	extern char * escape[];
	extern char songbuf[];
	extern char * sbp;
	char * start;
	char * tp;
	char h;

	/* fprintf(stderr, "get_stanza sbp=%d %c\n", sbp - songbuf - 1, *sbp); */

	if (*sbp == NULLCH)
		return "";

	/* extract the tag at the start (if there is one) */
	if (is_tag(sbp) != 0)
	{
		start = sbp;
		while (*sbp != ' ' && *sbp != '\n')
			sbp++;
		h = *sbp;
		*(sbp++) = NULLCH;
		/* fprintf(stderr, "tag %s, term = %d\n", start, h); */
		store_escape(ESCAPEt, start);

		/* Now the tag number (or other text) if any */
		if (h == ' ')
		{
			start = sbp;
			while (*sbp != '\n')
				sbp++;
			*(sbp++) = NULLCH;
			store_escape(ESCAPEn, start);
		}
		else
			store_escape(ESCAPEn, "");
	}
	else
	{
		store_escape(ESCAPEt, "");
		store_escape(ESCAPEn, "");
	}

	tp = temp;

	while (is_tag(sbp) == 0)
	{
		start = sbp;
		while (*sbp != '\n' && *sbp != NULLCH)
			sbp++;
		strncpy(tp, start, sbp - start);
		tp += sbp - start;
		strcpy(tp, escape[ESCAPE_line]);
		tp += 6;
		if (*sbp == NULLCH)
			break;
		sbp++;
	}

	/* we don't want a line break on the last one */
	tp -= 6;
	*tp = NULLCH;

	store_escape(ESCAPEz, temp);

	/* fprintf(stderr, "get_stanza returns %s\n", escape[ESCAPEz]); */

	return escape[ESCAPEz];
}

int is_tag(char * cp)
{
	int i, tl;

	/* fprintf(stderr, "is_tag %10.10s\n", cp); */
	for (i = 0; *tags[i] != NULLCH; i++)
	{
		tl = strlen(tags[i]);
		if (strncasecmp(cp, tags[i], tl) != 0)
			continue;
		if (*(cp + tl) == '\n')
			return i + 1;
		/* the next character must be a space */
		if (*(cp + tl) != ' ')
			continue;
		/* and the one after numeric or ( */
		if (!isdigit(*(cp + tl + 1)) && *(cp + tl + 1) != '(')
			continue;
	/* fprintf(stderr, "is_tag returns %d\n", i + 1); */
		return i + 1;
	}
	return 0;
}

void put_schedule(FILE * out, FILE * in)
{
	extern char * escape[];
	extern int inschd, insong;

	inschd = FALSE;
	insong = FALSE;

	/* fprintf(stderr, "put_schedule\n"); */

	put_text(out, escape[ESCAPE_schedule], in);
	return;
}

/* note - this subroutine is recursive. Care must be taken with any
	static variables that it uses directly or indirectly */
void put_text(FILE * out, char * string, FILE * in)
{
	extern char * escape[];
	extern int inschd, insong;

	int i;
	char c;
	char * cp;

	/* fprintf(stderr, "put_text(%s) %d %d\n", string, inschd, insong); */

	cp = string;
	for (; (c = *cp) != NULLCH; cp++)
	{
		if (c != '%')
			fputc(c, out);
		else
		{
			c = *(++cp);
			/* fprintf(stderr, "escape(%c)\n", c); */
			if (c == '%')
				fputc('%', out);
			else if (c == 'C' && !inschd)
			{ /* Schedule content escape */
				inschd=TRUE;
				while (strcmp(get_song(in), "") != 0)
				{
					/* PowerPoint, Images, Videos etc. have null content */
					if (*sbp != NULLCH)
						put_text(out, escape[ESCAPE_song], in);
				}
				inschd=FALSE;
			}
			else if (c == 's' && !insong)
			{ /* Song content escape */
				insong=TRUE;
				while (strcmp(get_stanza(), "") != 0)
				{
					/* see if we have a special template for this tag */
					for (i = 0; *tags[i] != NULLCH; i++)
					{
						if (strcasecmp(escape[ESCAPEt], tags[i]) == 0)
							break;
					}
					if (*tags[i] == NULLCH)
						i = 0;
					else
						i++;
					if (*(escape[ESCAPE_stanza + i]) == NULLCH)
						i = 0;
					put_text(out, escape[ESCAPE_stanza + i], in);
				}
				insong=FALSE;
			}
			else if (c > '0' && c <= 'z' && escape[c - '0'] != (char *)NULL)
				put_text(out, escape[c - '0'], in);
/* this has the effect of outputting null for any unknown or undefined
	escape characters. Beware of lone % characters in templates.
*/
		}
	}
	return;
}

char * store_escape(int n, char * string)
{
	int l;

	if (escape[n] != (char *)NULL)
		free(escape[n]);

	l = strlen(string);
	escape[n] = (char *)malloc(sizeof(char) * l + 1);
	strcpy(escape[n], string);

	return escape[n];
}
